/**
   Here are two recursive implementations of
   the edit distance dynamic programming problem.

                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
   Edit(s, t, i, j) = |       /
                      |      | 1 + Edit(s, t, i-1, j)
                      |  min | 1 + Edit(s, t, i, j-1)
                      |      |     Edit(s, t, i-1, j-1)  if s[i] == t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  if s[i] != t[j]
                       \      \



                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
                      |  Edit(s, t, i-1, j-1)              if s[i] == t[j]
   Edit(s, t, i, j) = |
                      |       /                         \
                      |      | 1 + Edit(s, t, i-1, j)    |
                      |  min | 1 + Edit(s, t, i, j-1)    | if s[i] != t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  }
                       \      \                         /

   NOTE: In the above pseudo-code, if a string has length n,
   then the character indices in the string are from 1 to n.
   But in Java, the character indices are from 0 to n-1.
*/
public class EditDistance
{
   /**
      Compute the edit distance between the two strings.

      @param s  source string
      @param t  target string
      @return the edit distance needed to transform s to t
   */
   public static int edit(String s, String t)
   {
      return edit1(s, t, s.length(), t.length());  // Use edit1() or edit2().
    //return edit2(s, t, s.length(), t.length());
   }


   /**
      Compute the edit distance between
      the prefix of each string.

      This is a purely recursive (and very inefficient,
      exponential) implementation of edit distance.

      @param s  source string
      @param t  target string
      @param i  length of the prefix from s
      @param j  length of the prefix from t
      @return the edit distance needed to transform s to t
   */
   public static int edit1(String s, String t, int i, int j)
   {
      if (i == 0) return j;
      if (j == 0) return i;

      final int del = 1 + edit1(s, t, i-1, j);   // recursion
      final int ins = 1 + edit1(s, t, i,   j-1); // recursion
            int rep = 1 + edit1(s, t, i-1, j-1); // recursion
      if ( s.charAt(i-1) == t.charAt(j-1) )
      {
         rep -= 1;
      }

      return Math.min(del, Math.min(ins, rep));
   }


   /**
      Compute the edit distance between
      the prefix of each string.

      This is a purely recursive (and very inefficient,
      exponential) implementation of edit distance.

      @param s  source string
      @param t  target string
      @param i  length of the prefix from s
      @param j  length of the prefix from t
      @return the edit distance needed to transform s to t
   */
   public static int edit2(String s, String t, int i, int j)
   {
      if (i == 0) return j;
      if (j == 0) return i;

      if ( s.charAt(i-1) == t.charAt(j-1) )
      {
         return edit2(s, t, i-1, j-1); // recursion
      }
      else
      {
         final int del = 1 + edit2(s, t, i-1, j);   // recursion
         final int ins = 1 + edit2(s, t, i,   j-1); // recursion
         final int rep = 1 + edit2(s, t, i-1, j-1); // recursion

         return Math.min(del, Math.min(ins, rep));
      }
   }



   public static void main(String[] args)
   {
      final String s1 = "food";
      final String t1 = "money";
      System.out.println( s1 + ", " + t1 + ", " + edit(s1, t1) ); // expect 4

      final String s2 = "algorithm";
      final String t2 = "altruistic";
      System.out.println( s2 + ", " + t2 + ", " + edit(s2, t2) ); // expect 6

      final String s3 = "snowy";
      final String t3 = "sunny";
      System.out.println( s3 + ", " + t3 + ", " + edit(s3, t3) ); // expect 3

      final String s4 = "exponential";
      final String t4 = "polynomial";
      System.out.println( s4 + ", " + t4 + ", " + edit(s4, t4) ); // expect 6
   }
}
