/**
   Here are two recursive implementations of
   the edit distance dynamic programming problem.

                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
   Edit(s, t, i, j) = |       /
                      |      | 1 + Edit(s, t, i-1, j)
                      |  min | 1 + Edit(s, t, i, j-1)
                      |      |     Edit(s, t, i-1, j-1)  if s[i] == t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  if s[i] != t[j]
                       \      \



                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
                      |  Edit(s, t, i-1, j-1)              if s[i] == t[j]
   Edit(s, t, i, j) = |
                      |       /                         \
                      |      | 1 + Edit(s, t, i-1, j)    |
                      |  min | 1 + Edit(s, t, i, j-1)    | if s[i] != t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  }
                       \      \                         /

   NOTE: In the above pseudo-code, if a string has length n,
   then the character indices in the string are from 1 to n.
   But in Java, the character indices are from 0 to n-1.
*/
public class EditDistance_v2
{
   /**
      Compute the edit distance between the two strings.

      @param s  source string
      @param t  target string
      @return a String that encodes the edits needed to transform s to t
   */
   public static String edit(String s, String t)
   {
      return edit1(s, t, s.length(), t.length());  // Use edit1() or edit2().
    //return edit2(s, t, s.length(), t.length());
   }


   /**
      Compute the edit distance between
      the prefix of each string.

      This is a purely recursive (and very inefficient,
      exponential) implementation of edit distance.

      @param s  source string
      @param t  target string
      @param i  length of the prefix from s
      @param j  length of the prefix from t
      @return a String that encodes the edits needed to transform s to t
   */
   public static String edit1(String s, String t, int i, int j)
   {
      if (i == 0) return "+".repeat(j);
      if (j == 0) return "-".repeat(i);

      final String del = edit1(s, t, i-1, j  ) + "-"; // recursion
      final String ins = edit1(s, t, i,   j-1) + "+"; // recursion
            String rep = edit1(s, t, i-1, j-1);       // recursion
      if ( s.charAt(i-1) == t.charAt(j-1) )
      {
         rep = rep + s.charAt(i-1);
      }
      else
      {
         rep = rep + "#";
      }

      final int delDist = decode(del);
      final int insDist = decode(ins);
      final int repDist = decode(rep);

      if (delDist <= insDist && delDist <= repDist)
         return del;
      else if (insDist <= repDist)
         return ins;
      else
         return rep;
   }


   /**
      Compute the edit distance between
      the prefix of each string.

      This is a purely recursive (and very inefficient,
      exponential) implementation of edit distance.

      @param s  source string
      @param t  target string
      @param i  length of the prefix from s
      @param j  length of the prefix from t
      @return a String that encodes the edits needed to transform s to t
   */
   public static String edit2(String s, String t, int i, int j)
   {
      if (i == 0) return "+".repeat(j);
      if (j == 0) return "-".repeat(i);

      if ( s.charAt(i-1) == t.charAt(j-1) )
      {
         return edit2(s, t, i-1, j-1) + s.charAt(i-1); // recursion
      }
      else
      {
         final String del = edit2(s, t, i-1, j  ) + "-"; // recursion
         final String ins = edit2(s, t, i,   j-1) + "+"; // recursion
         final String rep = edit2(s, t, i-1, j-1) + "#"; // recursion

         final int delDist = decode(del);
         final int insDist = decode(ins);
         final int repDist = decode(rep);

         if (delDist <= insDist && delDist <= repDist)
            return del;
         else if (insDist <= repDist)
            return ins;
         else
            return rep;
      }
   }


   public static int decode(String edits)
   {
      int editDistance = 0;
      for (int i = 0; i < edits.length(); ++i)
      {
         if ( edits.charAt(i) == '-'
           || edits.charAt(i) == '+'
           || edits.charAt(i) == '#' )
         {
            ++editDistance;
         }
      }
      return editDistance;
   }



   public static void main(String[] args)
   {
      final String s1 = "food";
      final String t1 = "money";
      final String edits1 = edit(s1, t1);
      System.out.println( s1 + ", " + t1 + ", " + edits1 + ", dist = " + decode(edits1) ); // expect 4

      final String s2 = "algorithm";
      final String t2 = "altruistic";
      final String edits2 = edit(s2, t2);
      System.out.println( s2 + ", " + t2 + ", " + edits2 + ", dist = " + decode(edits2) ); // expect 6

      final String s3 = "snowy";
      final String t3 = "sunny";
      final String edits3 = edit(s3, t3);
      System.out.println( s3 + ", " + t3 + ", " + edits3 + ", dist = " + decode(edits3) ); // expect 3

      final String s4 = "exponential";
      final String t4 = "polynomial";
      final String edits4 = edit(s4, t4);
      System.out.println( s4 + ", " + t4 + ", " + edits4 + ", dist = " + decode(edits4) ); // expect 6

      final String s5 = "pring";
      final String t5 = "print";
      final String edits5 = edit(s5, t5);
      System.out.println( s5 + ", " + t5 + ", " + edits5 + ", dist = " + decode(edits5) );

      final String s6 = "pring";
      final String t6 = "spring";
      final String edits6 = edit(s6, t6);
      System.out.println( s6 + ", " + t6 + ", " + edits6 + ", dist = " + decode(edits6) );

      final String s7 = "pring";
      final String t7 = "parking";
      final String edits7 = edit(s7, t7);
      System.out.println( s7 + ", " + t7 + ", " + edits7 + ", dist = " + decode(edits7) );
   }
}
