/**
   Here are two table driven implementations of
   the edit distance dynamic programming problem.

                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
   Edit(s, t, i, j) = |       /
                      |      | 1 + Edit(s, t, i-1, j)
                      |  min | 1 + Edit(s, t, i, j-1)
                      |      |     Edit(s, t, i-1, j-1)  if s[i] == t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  if s[i] != t[j]
                       \      \



                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
                      |  Edit(s, t, i-1, j-1)              if s[i] == t[j]
   Edit(s, t, i, j) = |
                      |       /                         \
                      |      | 1 + Edit(s, t, i-1, j)    |
                      |  min | 1 + Edit(s, t, i, j-1)    | if s[i] != t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  }
                       \      \                         /

   NOTE: In the above pseudo-code, if a string has length n,
   then the character indices in the string are from 1 to n.
   But in Java, the character indices are from 0 to n-1.
*/
public class EditDistance_v3
{
   public static boolean printTable = false;
   public static boolean decodeTable = false;

   /**
      Compute the edit distance between the two strings.

      @param s  source string
      @param t  target string
      @return the edit distance needed to transform s to t
   */
   public static int edit(String s, String t)
   {
      // The characters of the source are the rows of the table.
      // The characters of the target are the columns of the table.
      int[][] dpTable = new int[1 + s.length()][1 + t.length()];

      edit1(s, t, dpTable);  // Use edit1() or edit2().
    //edit2(s, t, dpTable);

      // Print the dynamic programming table.
      if (printTable)
      {
         System.out.printf("Here is the bottom-up dynamic programming table.\n");
         System.out.print("  ");
         for (int j = 0; j < t.length(); ++j)
         {
            System.out.print("  " + t.charAt(j));
         }
         System.out.println();
         for (int i = 0; i <= s.length(); ++i)
         {
            for (int j = 0; j <= t.length(); ++j)
            {
               System.out.printf("%2d ", dpTable[i][j]);
            }
            System.out.printf("\n");
         }
      }

      if (decodeTable)
      {
         System.out.println( decode(s, t, dpTable) );
         System.out.println( decodeR(s, t, s.length(), t.length(), dpTable) );
      }

      return dpTable[s.length()][t.length()];
   }


   /**
      Compute the edit distance between the two strings.

      This is a bottom-up, iterative, implementation of edit distance.

      @param s  source string
      @param t  target string
      @param dpTable  dynamic programming table to fill in
      @return the edit distance needed to transform s to t
   */
   public static void edit1(String s, String t, int[][] dpTable)
   {
      for (int i = 0; i <= s.length(); ++i) dpTable[i][0] = i;
      for (int j = 0; j <= t.length(); ++j) dpTable[0][j] = j;

      for (int i = 1; i <= s.length(); ++i)
      {
         for (int j = 1; j <= t.length(); ++j)
         {
            final int dif = (s.charAt(i-1) == t.charAt(j-1)) ? 0 : 1;
            final int del =   1 + dpTable[i - 1][j    ];
            final int ins =   1 + dpTable[i    ][j - 1];
            final int rep = dif + dpTable[i - 1][j - 1];

            dpTable[i][j] = Math.min(del, Math.min(ins, rep));
         }
      }
   }


   /**
      Compute the edit distance between the two strings.

      This is a bottom-up, iterative, implementation of edit distance.

      @param s  source string
      @param t  target string
      @param dpTable  dynamic programming table to fill in
      @return the edit distance needed to transform s to t
   */
   public static void edit2(String s, String t, int[][] dpTable)
   {
      for (int i = 0; i <= s.length(); ++i) dpTable[i][0] = i;
      for (int j = 0; j <= t.length(); ++j) dpTable[0][j] = j;

      for (int i = 1; i <= s.length(); ++i)
      {
         for (int j = 1; j <= t.length(); ++j)
         {
            if ( s.charAt(i-1) == t.charAt(j-1) )
            {
               dpTable[i][j] = dpTable[i-1][j-1];
            }
            else
            {
               final int del = 1 + dpTable[i - 1][j    ];
               final int ins = 1 + dpTable[i    ][j - 1];
               final int rep = 1 + dpTable[i - 1][j - 1];

               dpTable[i][j] = Math.min(del, Math.min(ins, rep));
            }
         }
      }
   }


   /**
      @param s  source string
      @param t  target string
      @param dpTable  filled in dynamic programming table
      @return a String that encodes the edits needed to transform s to t
   */
   public static String decode(String s, String t, int[][] dpTable)
   {
      String soln = "";

      int n = dpTable.length - 1;     // length of the source
      int m = dpTable[0].length - 1;  // length of the target

      while (0 != n || 0 != m)
      {
         if (n == 0) // source is empty
         {
            soln = "+" + soln; // insert into the source
            m = m - 1;
         }
         else if (m == 0) // target is empty
         {
            soln = "-" + soln; // delete from the source
            n = n - 1;
         }
         else if ( dpTable[n-1][m-1] <= dpTable[n-1][m]
                && dpTable[n-1][m-1] <= dpTable[n][m-1] )
         {
            if ( s.charAt(n-1) == t.charAt(m-1) )
            {
               soln = s.charAt(n-1) + soln;
            }
            else
            {
               soln = "#" + soln;
            }
            n = n - 1;
            m = m - 1;
         }
         else if ( dpTable[n-1][m] <= dpTable[n][m-1] )
         {
            soln = "-" + soln;
            n = n - 1;
         }
         else
         {
            soln = "+" + soln;
            m = m - 1;
         }
      }
      return soln;
   }


   /**
      This is a recursive version of the decode() method.

      @param s  source string
      @param t  target string
      @param i  length of the prefix from s
      @param j  length of the prefix from t
      @param dpTable  filled in dynamic programming table
      @return a String that encodes the edits needed to transform s to t
   */
   public static String decodeR(String s, String t, int i, int j, int[][] dpTable)
   {
      if (0 == i && 0 == j) return "";
      else if (0 == i) return decodeR(s, t, i, j-1, dpTable) + "+";
      else if (0 == j) return decodeR(s, t, i-1, j, dpTable) + "-";
      else if ( dpTable[i-1][j-1] <= dpTable[i-1][j]
             && dpTable[i-1][j-1] <= dpTable[i][j-1] )
      {
         if ( s.charAt(i-1) == t.charAt(j-1) )
            return decodeR(s, t, i-1, j-1, dpTable) + s.charAt(i-1);
         else
            return decodeR(s, t, i-1, j-1, dpTable) + "#";
      }
      else if ( dpTable[i-1][j] <= dpTable[i][j-1] )
         return decodeR(s, t, i-1, j, dpTable) + "-";
      else
         return decodeR(s, t, i, j-1, dpTable) + "+";
   }


   public static void main(String[] args)
   {
      printTable = true;

      final String s1 = "food";
      final String t1 = "money";
      System.out.println( s1 + ", " + t1 + ", " + edit(s1, t1) ); // expect 4

      final String s2 = "algorithm";
      final String t2 = "altruistic";
      System.out.println( s2 + ", " + t2 + ", " + edit(s2, t2) ); // expect 6

      final String s3 = "snowy";
      final String t3 = "sunny";
      System.out.println( s3 + ", " + t3 + ", " + edit(s3, t3) ); // expect 3

      final String s4 = "polynomial";
      final String t4 = "exponential";
      System.out.println( s4 + ", " + t4 + ", " + edit(s4, t4) ); // expect 6

      final String s5 = "cat";
      final String t5 = "hippopotamus";
      System.out.println( s5 + ", " + t5 + ", " + edit(s5, t5) );

      final String s6 = "hippopotamus";
      final String t6 = "cat";
      System.out.println( s6 + ", " + t6 + ", " + edit(s6, t6) );

      final String s7 = "abbb";
      final String t7 = "abb";
      System.out.println( s7 + ", " + t7 + ", " + edit(s7, t7) );

      final String s8 = "aaab";
      final String t8 = "aab";
      System.out.println( s8 + ", " + t8 + ", " + edit(s8, t8) );
   }
}
