/**
   Here are two table driven implementations of
   the edit distance dynamic programming problem.

                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
   Edit(s, t, i, j) = |       /
                      |      | 1 + Edit(s, t, i-1, j)
                      |  min | 1 + Edit(s, t, i, j-1)
                      |      |     Edit(s, t, i-1, j-1)  if s[i] == t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  if s[i] != t[j]
                       \      \



                       /
                      |  i    if j == 0
                      |  j    if i == 0
                      |
                      |  Edit(s, t, i-1, j-1)              if s[i] == t[j]
   Edit(s, t, i, j) = |
                      |       /                         \
                      |      | 1 + Edit(s, t, i-1, j)    |
                      |  min | 1 + Edit(s, t, i, j-1)    | if s[i] != t[j]
                      |      | 1 + Edit(s, t, i-1, j-1)  }
                       \      \                         /

   NOTE: In the above pseudo-code, if a string has length n,
   then the character indices in the string are from 1 to n.
   But in Java, the character indices are from 0 to n-1.
*/
public class EditDistance_v4
{
   public static boolean printTable = false;
   public static boolean decodeTable = false;

   /**
      Compute the edit distance between the two strings.

      @param s  source string
      @param t  target string
      @return the edit distance needed to transform s to t
   */
   public static int edit(String s, String t)
   {
      // The characters of the source are the rows of the table.
      // The characters of the target are the columns of the table.
      int[][] dpTable = new int[1 + s.length()][1 + t.length()];

      for (int i = 0; i <= s.length(); ++i)
         for (int j = 0; j <= t.length(); ++j)
            dpTable[i][j] = -1; // Flag that denotes this sub-problem hasn't been solved yet.

      edit1(s, t, s.length(), t.length(), dpTable);  // Use edit1() or edit2().
    //edit2(s, t, s.length(), t.length(), dpTable);

      // Print the dynamic programming table.
      if (printTable)
      {
         System.out.printf("Here is the top-down (memoization) dynamic programming table.\n");
         System.out.print("  ");
         for (int j = 0; j < t.length(); ++j)
         {
            System.out.print("  " + t.charAt(j));
         }
         System.out.println();
         for (int i = 0; i <= s.length(); ++i)
         {
            for (int j = 0; j <= t.length(); ++j)
            {
               System.out.printf("%2d ", dpTable[i][j]);
            }
            System.out.printf("\n");
         }
      }

      if (decodeTable)
      {
         System.out.println( EditDistance_v3.decode(s, t, dpTable) );
         System.out.println( EditDistance_v3.decodeR(s, t, s.length(), t.length(), dpTable) );
      }

      return dpTable[s.length()][t.length()];
   }


   /**
      Compute the edit distance between the two strings.

      This is a top-down, memoization, implementation of edit distance.

      @param s  source string
      @param t  target string
      @param i  length of the prefix from s
      @param j  length of the prefix from t
      @param dpTable  dynamic programming table to fill in
      @return the edit distance needed to transform s to t
   */
   public static int edit1(String s, String t, int i, int j, int[][] dpTable)
   {
      if ( dpTable[i][j] >= 0 )   // If the answer is already in the table,
         return dpTable[i][j];

      // Otherwise, do the recursion.
      final int distance;
      if (i == 0)
      {
         distance = j;
      }
      else if (j == 0)
      {
         distance = i;
      }
      else
      {
         final int del = 1 + edit1(s, t, i-1, j,  dpTable);  // recursion
         final int ins = 1 + edit1(s, t, i,   j-1, dpTable); // recursion
               int rep = 1 + edit1(s, t, i-1, j-1, dpTable); // recursion
         if ( s.charAt(i-1) == t.charAt(j-1) )
         {
            rep -= 1;
         }

         distance = Math.min(del, Math.min(ins, rep));
      }

      // Cache this result in the table.
      dpTable[i][j] = distance;
      return distance;
   }


   /**
      Compute the edit distance between the two strings.

      This is a top-down, memoization, implementation of edit distance.

      @param s  source string
      @param t  target string
      @param i  length of the prefix from s
      @param j  length of the prefix from t
      @param dpTable  dynamic programming table to fill in
      @return the edit distance needed to transform s to t
   */
   public static int edit2(String s, String t, int i, int j, int[][] dpTable)
   {
      if ( dpTable[i][j] >= 0 )   // If the answer is already in the table,
         return dpTable[i][j];

      // Otherwise, do the recursion.
      final int distance;
      if (i == 0)
      {
         distance = j;
      }
      else if (j == 0)
      {
         distance = i;
      }
      else if ( s.charAt(i-1) == t.charAt(j-1) )
      {
         distance = edit2(s, t, i-1, j-1, dpTable); // recursion
      }
      else
      {
         final int del = 1 + edit2(s, t, i-1, j,   dpTable); // recursion
         final int ins = 1 + edit2(s, t, i,   j-1, dpTable); // recursion
         final int rep = 1 + edit2(s, t, i-1, j-1, dpTable); // recursion

         distance =  Math.min(del, Math.min(ins, rep));
      }

      // Cache this result in the table.
      dpTable[i][j] = distance;
      return distance;
   }




   public static void main(String[] args)
   {
      printTable = true;

      final String s1 = "food";
      final String t1 = "money";
      System.out.println( s1 + ", " + t1 + ", " + edit(s1, t1) ); // expect 4

      final String s2 = "algorithm";
      final String t2 = "altruistic";
      System.out.println( s2 + ", " + t2 + ", " + edit(s2, t2) ); // expect 6

      final String s3 = "snowy";
      final String t3 = "sunny";
      System.out.println( s3 + ", " + t3 + ", " + edit(s3, t3) ); // expect 3

      final String s4 = "polynomial";
      final String t4 = "exponential";
      System.out.println( s4 + ", " + t4 + ", " + edit(s4, t4) ); // expect 6

      final String s5 = "cat";
      final String t5 = "hippopotamus";
      System.out.println( s5 + ", " + t5 + ", " + edit(s5, t5) );

      final String s6 = "hippopotamus";
      final String t6 = "cat";
      System.out.println( s6 + ", " + t6 + ", " + edit(s6, t6) );

      final String s7 = "abbb";
      final String t7 = "abb";
      System.out.println( s7 + ", " + t7 + ", " + edit(s7, t7) );

      final String s8 = "aaab";
      final String t8 = "aab";
      System.out.println( s8 + ", " + t8 + ", " + edit(s8, t8) );
   }
}
