/**
   The first few Fibonacci numbers are

   0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, ...

   If we let F(n) denote the n'th Fibonacci number, then these
   numbers are defined by the following recurrence relation.

     F(0) = 0
     F(1) = 1
     F(n) = F(n-1) + F(n-2) for n >= 2

   You can find a list of Fibonacci numbers at
      https://planetmath.org/listoffibonaccinumbers

   You can read about them at
      https://en.wikipedia.org/wiki/Fibonacci_sequence

   Below are three ways to compute F(n), the n'th Fibonacci number,
     1) regular recursion,
     2) bottom-up iteration,
     3) memoized (top-down) recursion.
*/
public class Fibonacci
{
   /**
     Here is the most basic way to compute the n'th
     Fibonacci number. We directly implement the
     recurrence relation as a recursive function.
   */
   public static long fib1(int n)
   {
      final long fn;
      if ( n == 0 )
         fn = 0;
      else if ( n == 1 )
         fn = 1;
      else
         fn = fib1(n-1) + fib1(n-2); // recursions
      return fn;
   }


   /**
      Here is a "bottom up dynamic programming"
      implementation of the n'th Fibonacci number.

      This method is NOT recursive.
   */
   public static long fib2(int n)
   {
      if (n == 0)
         return 0;

      final long[] dpTable = new long[n+1];
      dpTable[0] = 0;
      dpTable[1] = 1;
      for (int i = 2; i <= n; ++i)
      {
         dpTable[i] = dpTable[i-1] + dpTable[i-2]; // array look-ups (instead of recursion)
      }
      return dpTable[n];
   }


   /**
      Here is another "bottom up dynamic programming"
      implementation of the n'th Fibonacci number.
      This version does not need an n-element array
      to compute Fib(n). This version needs just
      three local variables.
   */
   public static long fib2a(int n)
   {
      long fi = -1;
      if ( n == 0 )
         fi = 0;
      else if ( n == 1 )
         fi = 1;
      else
      {
         long fi_minus_2 = 0;
         long fi_minus_1 = 1;
         for (int i = 2; i <= n; ++i)
         {
            fi = fi_minus_1 + fi_minus_2;
            fi_minus_2 = fi_minus_1;
            fi_minus_1 = fi;
         }
      }
      return fi;
   }


   /**
      Here is a "top down dynamic programming"
      implementation of the n'th Fibonacci number.
      This is also called "memoization dynamic programing".
   */
   public static long fib3(int n)
   {
      final long[] dpTable = new long[n+1];

      for (int i = 0; i <= n; ++i)
      {
         dpTable[i] = -1;  // Initialize the table with "empty" values.
      }

      return fib3_helper(n, dpTable);
   }

   private static long fib3_helper(int n, long[] dpTable)
   {
      final long fn;
      if ( n == 0 )
         fn = 0;
      else if ( n == 1 )
         fn = 1;
      else  // Compare this with the purely recursive verion, Fib1().
      {     // This is kind of a "short circuited" recursion.
         if (dpTable[n-1] == -1) dpTable[n-1] = fib3_helper(n-1, dpTable); // recursion
         if (dpTable[n-2] == -1) dpTable[n-2] = fib3_helper(n-2, dpTable); // recursion
         fn = dpTable[n-1] + dpTable[n-2];
      }
      return fn;
   }

   /**
     Here is a different way to write Fib3_helper().

     This does the memoization in a slightly different way.
     It "reverses" the order of the recursion and the memoization.
   */
   public static long fib3_helper2(int n, long[] dpTable)
   {
      if ( dpTable[n] != -1 ) return dpTable[n]; // Look in the table first.

      final long fn;
      if ( n == 0 )       // If the result is not in the table,
         fn = 0;          // then do the recursion.
      else if ( n == 1 )
         fn = 1;
      else
         fn = fib3_helper2(n-1, dpTable) + fib3_helper2(n-2, dpTable); // recursion

      dpTable[n] = fn; // Store the new result in the table before returning.
      return fn;
   }
}
