
/**
   Here are three implementations of
   the longest common subsequence
   dynamic programming problem.

                     /
                    | 0                            if i == 0
                    | 0                            if j == 0
                    | 1 + lcs(a, b, i-1, j-1)      if a[i] == b[j]
  lcs(a, b, i, j) = |      /                   \
                    |     | lcs(a, b, i-1, j)  |
                    | max |                    |   if a[i] != b[j]
                    |     | lcs(a, b, i, j-1)  |
                     \     \                  /
*/
public class LCS
{
   public static boolean printTable = false;

   /**
      Compute the longest common subsequence
      from the two input arrays.

      @param a1  array of integers
      @param a2  array of integers
      @return the length of the longest common subsequence
   */
   public static int lcs(int[] a1, int[] a2)
   {
      // Use either lcs1() or lcs2() or lcs3().
      return lcs1(a1, a2, a1.length, a2.length);  // recursion
    //return lcs2(a1, a2, a1.length, a2.length);  // bottom-up
    //return lcs3(a1, a2, a1.length, a2.length);  // top-down
   }


   /**
      Compute the longest common subsequence between
      the prefix of array a1 with length i and the
      prefix from array a2 with length j.

      Here is a purely recursive (and very inefficient,
      exponential) implementation of lcs(a1,a2,i,j).

      @param a1  array of integers
      @param a2  array of integers
      @param i   length of the prefix to test from a1
      @param j   length of the prefix to test from a2
      @return the length of the longest common subsequence
   */
   public static int lcs1(int[] a1, int[] a2, int i, int j)
   {
      final int length;
      if ( i == 0 || j == 0 )
         length = 0;
      else if ( a1[i-1] == a2[j-1] )
         length = 1 + lcs1(a1, a2, i-1, j-1); // recursion
      else
         length = Math.max(lcs1(a1, a2, i-1, j  ),  // recursion
                           lcs1(a1, a2, i,   j-1)); // recursion

      return length;
   }



   /**
      Compute the longest common subsequence between
      the prefix of array a1 with length i and the
      prefix from array a2 with length j.

      Here is a bottom-up dynamic programming
      implementation of lcs(a1,a2,i,j).

      @param a1  array of integers
      @param a2  array of integers
      @param i   length of the prefix to test from a1
      @param j   length of the prefix to test from a2
      @return the length of the longest common subsequence
   */
   public static int lcs2(int[] a1, int[] a2, int i, int j)
   {
      int[][] dpTable = new int[i+1][j+1];

      for (int col = 0; col <= j; ++col)  // Initialize top row.
         dpTable[0][col] = 0;

      for (int row = 0; row <= i; ++row)  // Initialize left most column.
         dpTable[row][0] = 0;

      for (int row = 1; row <= i; ++row)
      {
         for (int col = 1; col <= j; ++col)
         {
            if (a1[row-1] == a2[col-1] )
               dpTable[row][col] = 1 + dpTable[row-1][col-1]; // table loop-up
            else
               dpTable[row][col] = Math.max(dpTable[row-1][col  ],  // table loop-up
                                            dpTable[row  ][col-1]); // table loop-up
         }
      }

      // Print the dynamic programming table.
      if (printTable)
      {
         System.out.printf("Here is the LCS bottom-up dynamic programming table.\n");
         for (int row = 0; row <= i; ++row)
         {
            for(int col = 0; col <= j; ++col)
            {
               System.out.printf(" %2d", dpTable[row][col]);
            }
            System.out.printf("\n");
         }
      }

      return dpTable[i][j];
   }



   /**
      Compute the longest common subsequence between
      the prefix of array a1 with length i and the
      prefix from array a2 with length j.

      Here is a top-down dynamic programming
      implementation of lcs(a1,a2,i,j). This
      is also called "memoization".

      @param a1  array of integers
      @param a2  array of integers
      @param i   length of the prefix to test from a1
      @param j   length of the prefix to test from a2
      @return the length of the longest common subsequence
   */
   public static int lcs3(int[] a1, int[] a2, int i, int j)
   {
      int length;

      int[][] dpTable = new int[i+1][j+1];

      for (int row = 0; row <= i; ++row)
         for (int col = 0; col <= j; ++col)
            dpTable[row][col] = -1; // Flag that denotes this problem hasn't been solved yet.

      length = lcs3_helper(a1, a2, i, j, dpTable);

      // Print the dynamic programming table.
      if (printTable)
      {
         System.out.printf("Here is the LCS top-down dynamic programming table.\n");
         for (int row = 0; row <= i; ++row)
         {
            for(int col = 0; col <= j; ++col)
            {
               System.out.printf(" %2d", dpTable[row][col]);
            }
            System.out.printf("\n");
         }
      }

      return length;
   }

   /**
      @param a1  array of integers
      @param a2  array of integers
      @param i   length of the prefix to test from a1
      @param j   length of the prefix to test from a2
      @param dpTable  2-dim array to use as the memoization table
      @return the length of the longest common subsequence
   */
   public static int lcs3_helper(int[] a1, int[] a2, int i, int j, int[][] dpTable)
   {                               // Compare this with the purely recursive verion.
      if (dpTable[i][j] >= 0)      // This is kind of a "short circuited" recursion.
         return dpTable[i][j];     // If the answer is already in the table,
                                   // then use that value.
      final int length;            // Otherwise, do the recursion.
      if ( i == 0 || j == 0 )
         length = 0;
      else if ( a1[i-1] == a2[j-1] )
         length = 1 + lcs3_helper(a1, a2, i-1, j-1, dpTable); // recursion
      else
         length = Math.max(lcs3_helper(a1, a2, i-1, j,   dpTable),  // recursion
                           lcs3_helper(a1, a2, i,   j-1, dpTable)); // recursion

      dpTable[i][j] = length;  // Cache this result in the table.
      return length;
   }



   /**
      This method will randomly permute the elements
      of the array between index low and high.
   */
   public static void permutation(int[] a, int low, int high)
   {
      final java.util.Random generator = new java.util.Random();

      int upperBound = high;
      // The array from low to upperBound is the "jar" to
      // draw randomly from, and the array from upperBound
      // to high is the list of numbers that have already
      // been drawn out of the "jar".
      for (int i = low; i <= high; ++i)
      {
         // Pick a slot in the "jar" part of the array.
         final int randomIndex = low + generator.nextInt(upperBound-low+1);

         // Swap a[upperBound-1] with a[randomIndex]
         // (that is, move an item from the "jar" part
         // of the array to the list part of the array).
         final int temp = a[upperBound];
         a[upperBound] = a[randomIndex];
         a[randomIndex] = temp;

         // Lower the boundary of the "jar" part of the array.
         --upperBound;
      }
   }



   public static void main(String[] args)
   {
      final int START = 5;
      final int END   = 20;
      printTable = true;
      for (int n = START; n <= END; ++n)
      {
         int[] a = new int[n];
         int[] b = new int[n];

         // Initialize the arrays with the numbers for 1 to n.
         for (int j = 0; j < a.length; ++j) a[j] = j+1;
         for (int j = 0; j < b.length; ++j) b[j] = j+1;

         // The permutation() method will randomly
         // permute the contents of each array.
         permutation(a, 0, a.length - 1);
         permutation(b, 0, b.length - 1);

         // If n is small enough, print out the arrays.
         if (a.length <= 50)
         {
            System.out.println(java.util.Arrays.toString(a));
            System.out.println(java.util.Arrays.toString(b));
         }

         // Compute the length of the
         // longest common subsequence
         // in the suffled arrays.
         final int length = lcs(a, b);
         System.out.printf("lcs(%d) = %d.\n", a.length, length);
      }
   }
}
