
import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

/**
   Compute the structure of a solution to
   the longest common subsequence
   dynamic programming problem.

                     /
                    | 0                            if i == 0
                    | 0                            if j == 0
                    | 1 + lcs(a, b, i-1, j-1)      if a[i] == b[j]
  lcs(a, b, i, j) = |      /                   \
                    |     | lcs(a, b, i-1, j)  |
                    | max |                    |   if a[i] != b[j]
                    |     | lcs(a, b, i, j-1)  |
                     \     \                  /
*/
public class LCS_v2
{
   public static boolean printTable = false;

   /**
      Compute the structure of a solution to the
      longest common subsequence from the two
      input arrays.

      Encode the structure of a solution as a List
      of triples (a1[n], n, m) where a1[n] is a
      common entry from the two arraya that is in
      the longest common subsequence and n is that
      entry's index in array a1 and m is that entry's
      index in array a2.

      @param a1  array of integers
      @param a2  array of integers
      @return a List that encodes the entries in the longest common subsequence
   */
   public static List<List<Integer>> lcs(int[] a1, int[] a2)
   {
      // Use either lcs1() or lcs2() or lcs3().
      return lcs1(a1, a2, a1.length, a2.length);  // recursion
    //return lcs2(a1, a2, a1.length, a2.length);  // bottom-up
    //return lcs3(a1, a2, a1.length, a2.length);  // top-down
   }


   /**
      Compute the structure of a solution to the
      longest common subsequence between the prefix
      of array a1 with length i and the prefix from
      array a2 with length j.

      Here is a purely recursive (and very inefficient,
      exponential) implementation of lcs(a1,a2,i,j).

      @param a1  array of integers
      @param a2  array of integers
      @param i   length of the prefix to test from a1
      @param j   length of the prefix to test from a2
      @return a List that encodes the entries in the longest common subsequence
   */
   public static List<List<Integer>> lcs1(int[] a1, int[] a2, int i, int j)
   {
      final List<List<Integer>> soln;
      if ( i == 0 || j == 0 )
         soln = new ArrayList<>();
      else if ( a1[i-1] == a2[j-1] )
      {
         soln = lcs1(a1, a2, i-1, j-1);
         soln.add(Arrays.asList(new Integer[]{a1[i-1], i-1, j-1}));
      }
      else
      {
         final List<List<Integer>> soln1 = lcs1(a1, a2, i-1, j  );
         final List<List<Integer>> soln2 = lcs1(a1, a2, i,   j-1);

         if (soln1.size() >= soln2.size())
            soln = soln1;
         else
            soln = soln2;
      }

      return soln;
   }



   /**
      Compute the structure of a solution to the
      longest common subsequence between the prefix
      of array a1 with length i and the prefix from
      array a2 with length j.

      Here is a bottom-up dynamic programming
      implementation of lcs(a1,a2,i,j).

      After filling in the dynamic programming table,
      the table is analyzed to recover the structure
      of a solution. All the information needed to
      build a solution is in the dynamic programming
      table.

      @param a1  array of integers
      @param a2  array of integers
      @param i   length of the prefix to test from a1
      @param j   length of the prefix to test from a2
      @return a List that encodes the entries in the longest common subsequence
   */
   public static List<List<Integer>> lcs2(int[] a1, int[] a2, int i, int j)
   {
      int[][] dpTable = new int[i+1][j+1];

      for (int col = 0; col <= j; ++col)  // Initialize top row.
         dpTable[0][col] = 0;

      for (int row = 0; row <= i; ++row)  // Initialize left most column.
         dpTable[row][0] = 0;

      for (int row = 1; row <= i; ++row)
      {
         for (int col = 1; col <= j; ++col)
         {
            if (a1[row-1] == a2[col-1] )
               dpTable[row][col] = 1 + dpTable[row-1][col-1]; // table loop-up
            else
               dpTable[row][col] = Math.max(dpTable[row-1][col  ],  // table loop-up
                                            dpTable[row  ][col-1]); // table loop-up
         }
      }

      // Print the dynamic programming table.
      if (printTable)
      {
         System.out.printf("Here is the LCS bottom-up dynamic programming table.\n");
         for (int row = 0; row <= i; ++row)
         {
            for(int col = 0; col <= j; ++col)
            {
               System.out.printf(" %2d", dpTable[row][col]);
            }
            System.out.printf("\n");
         }
      }

      // Decode the dynamic programming table
      // into the structure of a solution.
      return decode(a1, a2, dpTable);
   }



   /**
      Compute the structure of a solution to the
      longest common subsequence between the prefix
      of array a1 with length i and the prefix from
      array a2 with length j.

      Here is a top-down dynamic programming
      implementation of lcs(a1,a2,i,j). This
      is also called "memoization".

      After filling in the dynamic programming table,
      the table is analyzed to recover the structure
      of a solution. All the information needed to
      build a solution is in the dynamic programming
      table.

      @param a1  array of integers
      @param a2  array of integers
      @param i   length of the prefix to test from a1
      @param j   length of the prefix to test from a2
      @return a List that encodes the entries in the longest common subsequence
   */
   public static List<List<Integer>> lcs3(int[] a1, int[] a2, int i, int j)
   {
      int[][] dpTable = new int[i+1][j+1];

      for (int row = 0; row <= i; ++row)
         for (int col = 0; col <= j; ++col)
            dpTable[row][col] = -1; // Flag that denotes this problem hasn't been solved yet.

      LCS.lcs3_helper(a1, a2, i, j, dpTable);

      // Print the dynamic programming table.
      if (printTable)
      {
         System.out.printf("Here is the LCS top-down dynamic programming table.\n");
         for (int row = 0; row <= i; ++row)
         {
            for(int col = 0; col <= j; ++col)
            {
               System.out.printf(" %2d", dpTable[row][col]);
            }
            System.out.printf("\n");
         }
      }

      // Decode the dynamic programming table
      // into the structure of a solution.
      return decode(a1, a2, dpTable);
   }


   /**
      @param a1  array of integers
      @param a2  array of integers
      @param dpTable  dynamic programming table encoding the solution to thee LCS problem
      @return a List that encodes the entries in the longest common subsequence
   */
   public static List<List<Integer>> decode(int[] a1, int[] a2, int[][] dpTable)
   {
      final List<List<Integer>> soln = new ArrayList<>();

      int m = dpTable[0].length - 1;
      int n = dpTable.length - 1;

      while (0 != n && 0 != m)
      {
         if (a1[n-1] == a2[m-1])
         {
            soln.add(0, Arrays.asList(new Integer[]{a1[n-1], n-1, m-1}));
            n = n - 1;
            m = m - 1;
         }
         else
         {
            if (dpTable[n][m] == dpTable[n-1][m])
            {
               n = n - 1;
            }
            else // dpTable[n][m] == dpTable[n][m-1]
            {
               m = m - 1;
            }
         }
      }
      return soln;
   }



   public static void main(String[] args)
   {
      printTable = true;

      int[] a1 = {3, 5, 2, 7, 8, 1, 2, 3, 9, 6, 4, 6};
      int[] a2 = {8, 6, 8, 2, 5, 1, 4, 8, 5, 3, 9, 0, 1, 2, 3, 5};
      System.out.println("a1 = " + java.util.Arrays.toString(a1));
      System.out.println("a2 = " + java.util.Arrays.toString(a2));
      System.out.println( "lcs(a1,a2) = " + lcs(a1, a2) );
      System.out.println();

      final int START = 5;
      final int END   = 20;
      for (int n = START; n <= END; ++n)
      {
         int[] a = new int[n];
         int[] b = new int[n];

         // Initialize the arrays with the numbers for 1 to n.
         for (int j = 0; j < a.length; ++j) a[j] = j+1;
         for (int j = 0; j < b.length; ++j) b[j] = j+1;

         // The permutation() method will randomly
         // permute the contents of each array.
         LCS.permutation(a, 0, a.length - 1);
         LCS.permutation(b, 0, b.length - 1);

         // If n is small enough, print out the arrays.
         if (a.length <= 50)
         {
            System.out.println(java.util.Arrays.toString(a));
            System.out.println(java.util.Arrays.toString(b));
         }

         // Compute the solution of the
         // longest common subsequence
         // in the suffled arrays.
         final List<List<Integer>> soln = lcs(a, b);
         System.out.printf("lcs(%d) = " + soln + ".\n", a.length);
      }

      System.out.println();
      int[] a = new int[40];
      int[] b = new int[40];
      for (int j = 0; j < a.length; ++j) a[j] = j+1;
      for (int j = 0; j < b.length; ++j) b[j] = j+1;
      LCS.permutation(a, 0, a.length - 1);
      LCS.permutation(b, 0, b.length - 1);
      System.out.println("a = " + java.util.Arrays.toString(a));
      System.out.println("b = " + java.util.Arrays.toString(b));
      System.out.println("lcs(a,b) = " + lcs(a, b));
   }
}
