
/**
   Here are three implementations of
   the longest increasing subsequence
   dynamic programming problem.

      lis(a) = max { LE(a, j) }
             0<=j<a.length
   where

      LE(a, j) = 1 + max { LE(a, i) }
                    i < j
                 a[i] < a[j]
*/
public class LIS
{
   public static boolean printTable = false;

   /**
      Compute the longest increasing subsequence
      using the helper function LE(a, j).

      @param a  array of integers
      @return the length of the longest increasing subsequence of a
   */
   public static int lis(int[] a)
   {
      int length = 0;

      for (int j = 0; j < a.length; ++j)
      {
         length = Math.max( length, LE3(a, j) );  // recursion
       //length = Math.max( length, LE2(a, j) );  // bottom-up
       //length = Math.max( length, LE3(a, j) );  // top-down
      }

      return length;
   }


   /**
      Compute the longest increasing subsequence
      from a that ends at entry a[j].

      This is a purely recursive (and very inefficient,
      exponential) implementation of LE(a, j).

      @param a  array of integers
      @param j  index of where the increasing subsequence ends
      @return the length of the longest increasing subsequence that ends at index j
   */
   public static int LE1(int[] a, int j)
   {
      int length = 0;
      if ( j == 0 )
      {
         length = 1;
      }
      else
      {
         for (int i = 0; i < j; ++i)
         {
            if ( a[i] < a[j] )
            {
               length = Math.max( length, LE1(a,i) ); // recursion
            }
         }
         length = 1 + length;
      }
      return length;
   }



   /**
      Compute the longest increasing subsequence
      from a that ends at entry a[j].

      This is a bottom-up dynamic programming
      implementation of LE(a, j).

      @param a  array of integers
      @param j  index of where the increasing subsequence ends
      @return the length of the longest increasing subsequence that ends at index j
   */
   public static int LE2(int[] a, int j)
   {
      int[] dpTable = new int[j+1];

      dpTable[0] = 1;
      for (int k = 1; k <= j; ++k)
      {
         int length = 0;
         for (int i = 0; i < k; ++i)
            if ( a[i] < a[k] )
               length = Math.max( length, dpTable[i] );
         dpTable[k] = 1+length;
      }

      // Print the dynamic programming tables.
      if (printTable)
      {
         if (j == 0)
            System.out.printf("Here are the LIS bottom-up dynamic programming tables.\n");
         for (int k = 0; k <= j; ++k)
         {
            System.out.printf("%2d ", dpTable[k]);
         }
         System.out.printf("\n");
      }

      return dpTable[j];
   }



   /**
      Compute the longest increasing subsequence
      from a that ends at entry a[j].

      This is a top-down dynamic programming
      implementation of LE(a, j). This is also
      called "memoization".

      @param a  array of integers
      @param j  index of where the increasing subsequence ends
      @return the length of the longest increasing subsequence that ends at index j
   */
   public static int LE3(int[] a, int j)
   {
      int length;

      int[] dpTable = new int[j+1];

      for (int i = 0; i <= j; ++i)
         dpTable[i] = -1; // Flag that denotes i'th problem hasn't been solved yet.

      length = LE3_helper(a, j, dpTable);

      // Print the dynamic programming tables.
      if (printTable)
      {
         if (j == 0)
            System.out.printf("Here are the LIS top-down dynamic programming tables.\n");
         for (int k = 0; k <= j; ++k)
            System.out.printf(" %2d", dpTable[k]);
         System.out.printf("\n");
      }

      return length;
   }

   /**
      @param a  array of integers
      @param j  index of where the increasing subsequence ends
      @param dpTable  1-dim array to use as the memoization table
      @return the length of the longest increasing subsequence that ends at index j
   */
   private static int LE3_helper(int[] a, int j, int[] dpTable)
   {                         // Compare this with the purely recursive verion.
      if ( dpTable[j] >= 0 ) // This is kind of a "short circuited" recursion.
         return dpTable[j];  // If the answer is already in the table,
                             // then use that value.
      int length = 0;        // Otherwise, do the recursion.
      if ( j == 0 )
      {
         length = 1;
      }
      else
      {
         for (int i = 0; i < j; ++i)
         {
            if ( a[i] < a[j] )
            {
               length = Math.max( length, LE3_helper(a, i, dpTable) ); // recursion
            }
         }
         length = 1 + length;
      }

      dpTable[j] = length;  // Cache this result in the table.
      return length;
   }



   /**
      This method will randomly permute the elements
      of the array between index low and high.
   */
   public static void permutation(int[] a, int low, int high)
   {
      final java.util.Random generator = new java.util.Random();

      int upperBound = high;
      // The array from low to upperBound is the "jar" to
      // draw randomly from, and the array from upperBound
      // to high is the list of numbers that have already
      // been drawn out of the "jar".
      for (int i = low; i <= high; ++i)
      {
         // Pick a slot in the "jar" part of the array.
         final int randomIndex = low + generator.nextInt(upperBound-low+1);

         // Swap a[upperBound-1] with a[randomIndex]
         // (that is, move an item from the "jar" part
         // of the array to the list part of the array).
         final int temp = a[upperBound];
         a[upperBound] = a[randomIndex];
         a[randomIndex] = temp;

         // Lower the boundary of the "jar" part of the array.
         --upperBound;
      }
   }



   public static void main(String[] args)
   {
      final int START = 1;
      final int END   = 20;
      printTable = true;
      for (int n = START; n <= END; ++n)
      {
         int[] a = new int[n];

         // Initialize the array with the numbers for 1 to n.
         for (int j = 0; j < a.length; ++j) a[j] = j+1;

         // The permutation() method will randomly
         // permute the contents of the array.
         permutation(a, 0, a.length - 1);

         // If n is small enough, print out the array.
         if (a.length <= 50)
         {
            System.out.println(java.util.Arrays.toString(a));
         }

         // Compute the length of the
         // longest increasing subsequence
         // in the suffled array.
         final int length = lis(a);
         System.out.printf("lis(%d) = %d.\n", a.length, length);
      }


      System.out.println();


      printTable = false;
      for (int n = 0; n < 200; ++n)
      {
         int[] a = new int[n];

         // Initialize the array with the numbers for 1 to n.
         for (int j = 0; j < a.length; ++j) a[j] = j+1;

         // The permutation() method will randomly
         // permute the contents of the array.
         permutation(a, 0, a.length - 1);

         // Compute the length of the
         // longest increasing subsequence
         // in the suffled array.
         final int length = lis(a);
         System.out.printf("lis(%d) = %d.\n", a.length, length);
      }
   }
}
